from libc.math cimport fabs, log, sin, cos, acos, asin, sinh, cosh, atan2, sqrt, M_PI
from cymath.astro.root cimport bracket_and_solve_root, CreateKeplerEquationBinding, KeplerEquationBinding, KeplerEquation
from cymath.vector3 cimport Vector3
from cymath.astro.astro_state cimport ECI

cdef inline double stumpff_c(double z):
    """
    Stumpff C
    From stumpC.m equation 3.50      
    Appendix D.4 Preliminary Orbit Determination, Orbital Mechanics for Engineering Students
    
    """
    cdef double c = 1.0 / 2.0
    if z > 0:
        c = (1.0 - cos(sqrt(z))) / z
    elif z < 0:
        c = (cosh(sqrt(-z)) - 1.0) / (-z)
        
    return c
    
cdef inline double stumpff_s(double z):
    """
    Stumpff S
    #From stumpS.m equation 3.49   
    Appendix D.4 Preliminary Orbit Determination, Orbital Mechanics for Engineering Students
    """
    cdef double s = 1.0 / 6.0
    cdef double sqrtz = 0.0
    if(z > 0):
        sqrtz = sqrt(z) 
        s = (sqrtz - sin(sqrtz)) / (sqrtz**3.0)
    elif(z < 0):
        sqrtz = sqrt(-z)
        s = (sinh(sqrtz) - sqrtz) / (sqrtz**3.0)

    return s   
    
cdef inline double kepler_e(double e, double m, double eccentricity):
    """
    Eccentric Anomaly (e)
    """
    return e - eccentricity * sin(e) - m
    
cdef inline double kepler_de(double de, double dm, double sigma0, double sqrta, double a, double r):
    """
    Eccentric Anomaly Difference (de)
    """
    return (-dm + de + sigma0 / sqrta * (1.0 - cos(de)) - (1.0 - r / a) * sin(de))

cdef inline double kepler_dh(double dh, double dn, double sigma0, double sqrta, double a, double r):
    """
    Hyperbolic Anomaly Difference (dh)
    """
    return (-dn - dh + sigma0 / sqrta * (cosh(dh) - 1.0) + (1.0 - r / a) * sinh(dh))
    
cdef inline double kepler_ds(double ds, double dt, double r0, double v0, double alpha, double mu):
    """
    From pykep
    Universal Anomaly Difference (ds)
    """
    cdef double s = stumpff_s(alpha * ds**2.0)
    cdef double c = stumpff_c(alpha * ds**2.0)
    return -sqrt(mu) * dt + r0 * v0 * ds**2.0 * c / sqrt(mu) + (1.0 - alpha * r0) * ds**3.0 * s + r0 * ds
    
#cdef int ASTRO_ITERS = 500
  
cpdef object propagate_lagrangian(Vector3 r, Vector3 v, double t, double mu)
cpdef ECI propagate_lagrangian_u(ECI state, double deltat_sec)